# *********************************************
#  命令語インタプリタ Version 1.0 (2020/05/06)
# *********************************************
# このコードは無償・無保証です。自由にお使いください。
#
# [仕様]
# LDCv2!が出力する命令語プログラムをRaspberry Pi Zero上で実行するインタプリタで、Pythonで記述してあります。
# LDCv2!で作成したラダー図を命令語変換したファイル(plc.il)を読み込み、命令語プログラムを繰り返し実行(スキャン)します。
# 毎スキャン前後に、LD Cv2!のメモリ(M)とRaspberry PiのGPIO間で、次のデータ転送を行なっています。
#  スキャン前 GPIO(20),GPIO(21),GPIO(22),GPIO(23) --> M20,M21,M22,M23
#  スキャン後 M24,M25,M26,M27 --> GPIO(24),GPIO(25),GPIO(26),GPIO(27)
# 詳細情報は、https://www.takami.com/ldcv/から文書「LDCv2!のROM化開発手順」の第5版以上をダウンロードください。

#--------------------------------------------------
# Raspberry Pi Zero初期設定
#--------------------------------------------------
import RPi.GPIO as GPIO
GPIO.setwarnings(False) # ((以前のポート設定と競合しても警告メッセージを出しません))
GPIO.setmode(GPIO.BCM)
 
#GPIO20～GPIO23を入力ポートに設定
# ((必要に応じて書き換えてください))
GPIO.setup(20,GPIO.IN)
GPIO.setup(21,GPIO.IN)
GPIO.setup(22,GPIO.IN)
GPIO.setup(23,GPIO.IN)

#GPIO24～GPIO27を出力ポートに設定
# ((必要に応じて書き換えてください))
GPIO.setup(24,GPIO.OUT)
GPIO.setup(25,GPIO.OUT)
GPIO.setup(26,GPIO.OUT)
GPIO.setup(27,GPIO.OUT)

#--------------------------------------------------
# 命令語プログラム(plc.il)読込、語(トークン)に分解
#--------------------------------------------------
# ((プログラムにシフトJISコードが混じっていると、ダメみたいです))
f = open("plc.il","r")
lines = f.readlines()
f.close

words=list()
for i in range(len(lines)):
    if lines[i].startswith(" "): # 先頭が空白の行は無視
        pass
    elif lines[i].startswith("("): # 先頭が括弧の行は無視
        pass
    else:
        words += lines[i].split() # 語(トークン)に分解
        
print(words) # 分解したプログラムをコンソールに表示

#--------------------------------------------------
# ＰＬＣ初期化
#--------------------------------------------------
#エリア宣言と初期設定
M = [int(0)]*0x800 # Ｉ／ＯメモリＭ
S = [int(0)]*0x80  # 特殊メモリＳ
T = [int(0)]*0x80  # タイマＴ
C = [int(0)]*0x80  # カウンタＣ

pc = int(0) # プログラムカウンタ
rg = int(0x00) # 演算レジスタ
mc = int(0x01) # マスタコントロールレジスタ
stk = int(0x00) # 演算スタック
tk = [int(0)]*0x80 # タイマ経過値
ck = [int(0)]*0x80 # カウンタ経過

cf = [int(0x00)]*0x80 # カウンタクリアフラグ 
cl = [int(0x00)]*0x80 # カウンタ作業変数 
s0_old = int(0x00) # Ｓ０作業変数
s0_new = int(0x00) # Ｓ０作業変数
cnt_10 = int(0) # Ｓ１１作業変数
cnt_100 = int(0) # Ｓ１２作業変数
cnt_1000 = int(0) # Ｓ１３作業変数
cnt_10000 = int(0) # Ｓ１４作業変数

#--------------------------------------------------
# ＰＬＣ運転開始
#--------------------------------------------------
print("\nPLC START!")

#運転開始１スキャン目の設定
S[0x1e] = 0x01 #S1Eの1スキャン目=1
S[0x1f] = 0x01 #S1Fの1スキャン目=1

while True: # 以下、繰り返し実行(外側ループ)
# ((このループの中に、迂闊にprint文を書くとスキャンが一瞬止まるので厳禁です。用途にもよるけど・・・))
    
#--------------------------------------------------
# 入力リフレッシュ
#--------------------------------------------------
# 入力ポートGPIO4～7の値を、仮想PLCのM4～M7に書込
# ((必要に応じて書き換えてください))
    M[0x20] = GPIO.input(20) 
    M[0x21] = GPIO.input(21) 
    M[0x22] = GPIO.input(22) 
    M[0x23] = GPIO.input(23) 

#--------------------------------------------------
# １スキャン実行 (プログラム先頭～末尾)
#--------------------------------------------------
# (1)毎スキャン先頭の処理
    pc =0
    rg = 0x00
    mc = 0x01
    stk = 0x00
    s0_old = s0_new
    s0_new = S[0]

# (2)命令語プログラムを先頭から１命令ずつを読出・実行(ＥＮＤ命令まで)
    while True: # 以下、繰り返し実行(内側ループ)

# (3)命令の種類により、プログラム(words)を先読み
# ((ここはもっとスッキリ書きたかった所))
        opc = words[pc]                                         # オペコード(LDなど)を読出
        # LD系命令のとき
        if (opc == "LD" or opc == "LDNOT" or opc == "AND" or opc == "ANDNOT" or opc == "OR" or opc == "ORNOT"):
            nam_str = words[pc+1][0:1] # 第1オペランド(名前部)の文字列 (例)"M"
            num_str = words[pc+1][1:] # 第1オペランド(番号部)の文字列 (例)"07F"
            num_int = int(num_str,base=16) # 第1オペランド(番号部)の数値化 (例)127 
            if nam_str == "M":
                opr1 = M[num_int]
            elif nam_str == "S":
                opr1 = S[num_int]
            elif nam_str == "T":
                opr1 = T[num_int]
            elif nam_str == "C":
                opr1 = C[num_int]
        # OUT系命令のとき
        elif (opc == "OUT" or opc == "SET" or opc == "RES" or opc == "CLRC"):
            nam_str = words[pc+1][0:1] # 第1オペランド(名前部)の文字列 (例)"M"
            num_str = words[pc+1][1:] # 第1オペランド(番号部)の文字列 (例)"07F"
            num_int = int(num_str,base=16) # 第1オペランド(番号部)の数値化 (例)127
        # OUTT系命令のとき           
        elif (opc == "OUTT" or opc == "OUTC"):                            
            nam_str = words[pc+1][0:1] # 第1オペランド(名前部)の文字列 (例)"T"
            num_str = words[pc+1][1:] # 第1オペランド(番号部)の文字列 (例)"01A"
            num_int = int(num_str,base=16) # 第1オペランド(番号部)の数値化 (例)26
            sv_int =  int(words[pc+2],base=10) # 第2オペランド(設定値)の数値化 (例)100
        # その他の命令のとき
        else:
            pass
        
# (4)LD系命令の演算処理
# ((マスタコントロールの仕組みをOUT命令側に入れたので、LD命令側では何もしてません！))
        ### LD ###
        if opc == "LD":
            stk = stk << 1
            stk = stk | (rg & 0x01)
            rg = opr1
            pc+=2

        ### LDNOT ###
        elif opc == "LDNOT":
            stk = stk << 1
            stk = stk | (rg & 0x01)
            rg = ~opr1
            pc+=2

        ### AND ###
        elif opc == "AND":
            rg = rg & opr1
            pc+=2

        ### ANDNOT ###
        elif opc == "ANDNOT":
            rg = rg & ~opr1
            pc+=2

        ### OR ###
        elif opc == "OR":
            rg = rg | opr1
            pc+=2

        ### ORNOT ###
        elif opc == "ORNOT":
            rg = rg | ~opr1
            pc+=2
            
# (5)OUT系命令の演算処理
        ### OUT ###
        elif opc == "OUT":
            if nam_str == "M":
                M[num_int] = rg & mc & 0x01
            elif nam_str == "S":
                S[num_int] = rg & mc & 0x01
            pc+=2
            
        ### SET ###
        elif opc == "SET":
            if (rg & mc &0x01) == 0x01:
                if nam_str == "M":
                    M[num_int] = 0x01
                elif nam_str == "S":
                    S[num_int] = 0x01
            pc+=2

        ### RES ###
        elif opc == "RES":
            if (rg & mc &0x01) == 0x01:
                if nam_str == "M":
                    [num_int] = 0x01
                elif nam_str == "S":
                    S[num_int] = 0x01
            pc+=2

        ### CLRC ###
        elif opc == "CLRC":
            if (rg & mc &0x01) == 0x01:
                cl[num_int] = rg & mc &0x01
            pc+=2

# (6)OUTT系命令の演算処理
        ### OUTT ###
        elif opc == "OUTT":
            if (rg & mc & 0x01) == 0x01: # 励磁
                if  ((s0_old ^ s0_new) & 0x01) == 0x01: # S0変化
                    if tk[num_int] != 65535: # 経過値≠65535
                        tk[num_int]+=1 # 経過値を+1
                if tk[num_int] >= sv_int: # 経過値>= 設定値
                    T[num_int] = 0x01
                else:
                    T[num_int] = 0x00
            else:
                T[num_int] = 0x00
                tk[num_int] = 0x00
            pc+=3

        ### OUTC ###
        elif opc == "OUTC":
            if ((cl[num_int] & 0x01) == 0x00): # クリア入力なし
                if (((cf[num_int] & 0x01) == 0x00) & ((rg & mc & 0x01) == 0x01)): # カウント入力(↑)
                    if ck[num_int] != 65535: # 経過値≠65535
                        ck[num_int]+=1 # 経過値を+1
            else: # クリア入力あり
                ck[num_int] = 0 # 経過値をクリア
            #　(ここまでで経過値を更新)
            if ck[num_int] >= sv: # 経過値>= 設定値
                C[num_int] = 0x01 # ON
            else:
                C[num_int] = 0x00 # OFF
            cf[num_int] = rg & mc # カウント入力を記憶(次回スキャンのため)
            pc+=3

# (7)その他の命令の演算処理
        ### ANDSTACK ###
        elif opc == "ANDSTACK":
            rg = rg & (stk & 0x01)
            stk = stk >> 1
            pc+=1

        ### ORSTACK ###
        elif opc == "ORSTACK":
            rg = rg | (stk & 0x01)
            stk = stk >> 1
            pc+=1
            
        ### MCS ###
        elif opc == "MCS":
            mc = rg & mc # ((ここのmcは冗長だったかな？))
            pc+=1
           
        ### MCR ###
        elif opc == "MCR":
            mc = 0x01
            pc+=1

        ### END ###
        elif opc == "END":
            S[0x10] = ~S[0x10] & 0x01
            cnt_10+=1
            cnt_10 = cnt_10 % 10
            cnt_100+=1
            cnt_100 = cnt_100 % 100
            cnt_1000+=1
            cnt_1000 = cnt_1000 % 1000
            cnt_10000+=1
            cnt_10000 = cnt_10000 % 10000

            if cnt_10 == 0:
                S[0x11] = ~S[0x11] & 0x01
            if cnt_100 == 0:
                S[0x12] = ~S[0x12] & 0x01
            if cnt_1000 == 0:
                S[0x13] = ~S[0x13] & 0x01
            if cnt_10000 == 0:
                S[0x14] = ~S[0x14] & 0x01
            S[0x1e] = 0x00
            S[0x1f] = 0x01
            break # １スキャン終了(内側ループを抜ける)
        
 # (8)解釈不能のとき       
        else:
            print("\nERROR!")
            exit()

#--------------------------------------------------
# 出力リフレッシュ
#--------------------------------------------------
# 仮想PLCのM4～M7の値を、出力ポートGPIO4～7に書込
# ((必要に応じて書き換えてください))
    GPIO.output(24,M[0x24])
    GPIO.output(25,M[0x25])
    GPIO.output(26,M[0x26])
    GPIO.output(27,M[0x27]) 

# 入力リフレッシュに戻る(外側ループを繰り返す)
